{{ current_settings }}


# --- Settings for Fly.io. ---
import os, dj_database_url

if os.environ.get("ON_FLYIO"):
    # Use secret, if set, to update DEBUG value.
    if os.environ.get("DEBUG") == "FALSE":
        DEBUG = False
    elif os.environ.get("DEBUG") == "TRUE":
        DEBUG = True

    # Set a Fly.io-specific allowed host.
    # ALLOWED_HOSTS.append('{{ deployed_project_name }}.fly.dev')
    # Using '*' while further troubleshooting CSRF issue.
    ALLOWED_HOSTS.append('*')

    # Use the Fly.io Postgres database.
    db_url = os.environ.get("DATABASE_URL")
    DATABASES['default'] = dj_database_url.parse(db_url)



{{ current_settings }}


# platform.sh settings
from platformshconfig import Config
import os

# Import some Platform.sh settings from the environment.
config = Config()
if config.is_valid_platform():

    ALLOWED_HOSTS.append('*')
    DEBUG = False

    STATIC_URL = '/static/'

    if config.appDir:
        STATIC_ROOT = os.path.join(config.appDir, 'static')
    if config.projectEntropy:
        SECRET_KEY = config.projectEntropy

    if not config.in_build():
        db_settings = config.credentials('database')
        DATABASES = {
            'default': {
                'ENGINE': 'django.db.backends.postgresql',
                'NAME': db_settings['path'],
                'USER': db_settings['username'],
                'PASSWORD': db_settings['password'],
                'HOST': db_settings['host'],
                'PORT': db_settings['port'],
            },
            'sqlite': {
                'ENGINE': 'django.db.backends.sqlite3',
                'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
            }
        }


"""Manages all platform.sh-specific aspects of the deployment process."""

# Note: All public-facing references to platform.sh will include a dot, dash, or
#  underscore, ie platform_sh.
#  Internally, we won't use a space, ie platformsh or plsh.

import sys, os, re, subprocess
from pathlib import Path

from django.conf import settings
from django.core.management.base import CommandError
from django.core.management.utils import get_random_secret_key
from django.utils.crypto import get_random_string
from django.template.engine import Engine
from django.template.loaders.app_directories import Loader
from django.template.loader import render_to_string
from django.utils.safestring import mark_safe

from simple_deploy.management.commands.utils import deploy_messages as d_msgs
from simple_deploy.management.commands.utils import deploy_messages_platformsh as plsh_msgs


class PlatformshDeployer:
    """Perform the initial deployment of a simple project.
    Configure as much as possible automatically.
    """

    def __init__(self, command):
        """Establishes connection to existing simple_deploy command object."""
        self.sd = command
        self.stdout = self.sd.stdout


    def deploy(self, *args, **options):
        self.sd.write_output("Configuring project for deployment to Platform.sh...")

        self._add_platformsh_settings()

        # DEV: Group this with later yaml generation methods.
        self._generate_platform_app_yaml()

        # DEV: These can be done in one pass.
        self._add_platformshconfig()
        self._add_gunicorn()
        self._add_psycopg2()

        # DEV: These could be refactored.
        self._make_platform_dir()
        self._generate_routes_yaml()
        self._generate_services_yaml()

        self._conclude_automate_all()

        self._show_success_message()


    # --- Methods used in this class ---

    def _add_platformsh_settings(self):
        """Add platformsh-specific settings."""
        # The only project-specific setting is the ALLOWED_HOSTS; that makes
        #   modifying settings *much* easier than for other platforms.
        #   Just check if the settings are present, and if not, dump them in.

        # DEV: Modify this to make a more specific ALLOWED_HOSTS entry.
        #   For now, at proof of concept stage, it's just '*'.

        self.sd.write_output("\n  Checking if platform.sh-specific settings present in settings.py...")

        with open(self.sd.settings_path) as f:
            settings_string = f.read()

        if 'if config.is_valid_platform():' in settings_string:
            self.sd.write_output("\n    Found platform.sh settings block in settings.py.")
            return

        # Add platformsh settings block.
        # This comes from a template, which will make it easier to modify things
        #   like ALLOWED_HOSTS. This approach may work better for other platforms
        #   as well.
        self.sd.write_output("    No platform.sh settings found in settings.py; adding settings...")
        my_loader = Loader(Engine.get_default())
        my_template = my_loader.get_template('platformsh_settings.py')

        # Build context dict for template.
        safe_settings_string = mark_safe(settings_string)
        context = {'current_settings': safe_settings_string}
        template_string = render_to_string('platformsh_settings.py', context)

        path = Path(self.sd.settings_path)
        path.write_text(template_string)

        msg = f"    Modified settings.py file: {path}"
        self.sd.write_output(msg)


    def _get_platformsh_settings(self):
        """Get any platformsh-specific settings that are already in place.
        """
        # If any platformsh settings have already been written, we don't want to
        #  add them again. This assumes a section at the end, starting with a
        #  check for `if config.is_valid_platform():`

        with open(self.sd.settings_path) as f:
            settings_lines = f.readlines()

        self.found_platformsh_settings = False
        self.current_platformsh_settings_lines = []
        for line in settings_lines:
            if "if config.is_valid_platform():" in line:
                self.found_platformsh_settings = True
            if self.found_platformsh_settings:
                self.current_platformsh_settings_lines.append(line)


    def _generate_platform_app_yaml(self):
        """Create .platform.app.yaml file, if not present."""

        # File should be in project root, if present.
        self.sd.write_output(f"\n  Looking in {self.sd.git_path} for .platform.app.yaml file...")
        p_app_yaml_present = '.platform.app.yaml' in os.listdir(self.sd.git_path)

        if p_app_yaml_present:
            self.sd.write_output("    Found existing .platform.app.yaml file.")
        else:
            # Generate file from template.
            self.sd.write_output("    No .platform.app.yaml file found. Generating file...")
            my_loader = Loader(Engine.get_default())
            my_template = my_loader.get_template('platform.app.yaml')

            # Build context dict for template.
            context = {
                'project_name': self.sd.project_name, 
                'deployed_project_name': self.deployed_project_name
                }
            template_string = render_to_string('platform.app.yaml', context)

            path = self.sd.project_root / '.platform.app.yaml'
            path.write_text(template_string)

            msg = f"\n    Generated .platform.app.yaml file: {path}"
            self.sd.write_output(msg)
            return path


    def _add_platformshconfig(self):
        """Add platformshconfig to project requirements."""
        self.sd.write_output("\n  Looking for platformshconfig...")

        if self.sd.using_req_txt:
            self.sd.add_req_txt_pkg('platformshconfig')
        elif self.sd.using_pipenv:
            self.sd.add_pipenv_pkg('platformshconfig')


    def _add_gunicorn(self):
        """Add gunicorn to project requirements."""
        self.sd.write_output("\n  Looking for gunicorn...")

        if self.sd.using_req_txt:
            self.sd.add_req_txt_pkg('gunicorn')
        elif self.sd.using_pipenv:
            self.sd.add_pipenv_pkg('gunicorn')


    def _add_psycopg2(self):
        """Add psycopg2 to project requirements."""
        self.sd.write_output("\n  Looking for psycopg2...")

        if self.sd.using_req_txt:
            self.sd.add_req_txt_pkg('psycopg2')
        elif self.sd.using_pipenv:
            self.sd.add_pipenv_pkg('psycopg2')


    def _check_allowed_hosts(self):
        """Make sure project can be served from platformsh."""
        # This method is specific to platformsh.

        self.sd.write_output("\n  Making sure project can be served from platform.sh...")

        # DEV: Configure an ALLOWED_HOSTS entry that's specific to this deployment.
        # Use '*' for now, to focus on more specific aspects of platformsh deployment.
        platformsh_host = '*'

        if platformsh_host in settings.ALLOWED_HOSTS:
            self.sd.write_output(f"    Found {platformsh_host} in ALLOWED_HOSTS.")
        else:
            new_setting = f"ALLOWED_HOSTS.append('{platformsh_host}')"
            msg_added = f"    Added {platformsh_host} to ALLOWED_HOSTS for the deployed project."
            msg_already_set = f"    Found {platformsh_host} in ALLOWED_HOSTS for the deployed project."
            self._add_platformsh_setting(new_setting, msg_added, msg_already_set)


    def _make_platform_dir(self):
        """Add a .platform directory, if it doesn't already exist."""

        # Directory should be in project root, if present.
        self.sd.write_output(f"\n  Looking in {self.sd.git_path} for .platform/ directory...")

        self.platform_dir_path = self.sd.git_path / '.platform'
        if self.platform_dir_path.exists():
            self.sd.write_output("    Found existing .platform/ directory.")
        else:
            self.platform_dir_path.mkdir()
            self.sd.write_output(f"    Made .platform directory: {self.platform_dir_path}")


    def _generate_routes_yaml(self):
        """Generate the .platform/routes.yaml file, if not present."""

        # File should be in self.platform_dir_path, if present.
        self.sd.write_output(f"\n  Looking in {self.platform_dir_path} for routes.yaml file...")
        routes_yaml_present = 'routes.yaml' in os.listdir(self.platform_dir_path)

        if routes_yaml_present:
            self.sd.write_output("    Found existing routes.yaml file.")
        else:
            # Generate file from template.
            self.sd.write_output("    No routes.yaml file found. Generating file...")
            my_loader = Loader(Engine.get_default())
            my_template = my_loader.get_template('routes.yaml')

            # Build context dict for template.
            context = {'deployed_project_name': self.deployed_project_name}
            template_string = render_to_string('routes.yaml', context)

            path = self.platform_dir_path / 'routes.yaml'
            path.write_text(template_string)

            msg = f"\n    Generated routes.yaml file: {path}"
            self.sd.write_output(msg)
            return path


    def _generate_services_yaml(self):
        """Generate the .platform/services.yaml file, if not present."""
        
        # File should be in self.platform_dir_path, if present.
        self.sd.write_output(f"\n  Looking in {self.platform_dir_path} for services.yaml file...")
        routes_yaml_present = 'services.yaml' in os.listdir(self.platform_dir_path)

        if routes_yaml_present:
            self.sd.write_output("    Found existing services.yaml file.")
        else:
            # Generate file from template.
            # DEV: We're not modifying this file, so it can just be copied 
            #   from templates file to .platform/ directory.
            self.sd.write_output("    No services.yaml file found. Generating file...")
            my_loader = Loader(Engine.get_default())
            my_template = my_loader.get_template('services.yaml')

            template_string = render_to_string('services.yaml')

            path = self.platform_dir_path / 'services.yaml'
            path.write_text(template_string)

            msg = f"\n    Generated services.yaml file: {path}"
            self.sd.write_output(msg)
            return path


    def _conclude_automate_all(self):
        """Finish automating the push to Platform.sh.
        - Commit all changes.
        - Call `platform push`.
        - Open project.
        """
        # Making this check here lets deploy() be cleaner.
        if not self.sd.automate_all:
            return

        self.sd.commit_changes()

        # Push project.
        self.sd.write_output("  Pushing to Platform.sh...")
        cmd = "platform push --yes"
        output = self.sd.execute_subp_run(cmd)
        self.sd.write_output(output)

        # Open project.
        self.sd.write_output("  Opening deployed app in a new browser tab...")
        cmd = "platform url --yes"
        output = self.sd.execute_subp_run(cmd)
        self.sd.write_output(output)

        # Get url of deployed project.
        #   This can be done with an re, but there's one line of output with
        #   a url, so finding that line is simpler.
        self.deployed_url = ''
        for line in output.stdout.decode().split('\n'):
            if 'https' in line:
                self.deployed_url = line.strip()


    def _show_success_message(self):
        """After a successful run, show a message about what to do next."""

        # DEV:
        # - Mention that this script should not need to be run again, unless
        #   creating a new deployment.
        #   - Describe ongoing approach of commit, push, migrate. Lots to consider
        #     when doing this on production app with users, make sure you learn.

        if self.sd.automate_all:
            msg = plsh_msgs.success_msg_automate_all(self.deployed_url)
            self.sd.write_output(msg)
        else:
            msg = plsh_msgs.success_msg(self.sd.log_output)
            self.sd.write_output(msg)


    # --- Methods called from simple_deploy.py ---

    def confirm_preliminary(self):
        """Deployment to platform.sh is in a preliminary state, and we need to be
        explicit about that.
        """
        # Skip this confirmation when unit testing.
        if self.sd.unit_testing:
            return

        self.stdout.write(plsh_msgs.confirm_preliminary)
        confirmed = self.sd.get_confirmation(skip_logging=True)

        if confirmed:
            self.stdout.write("  Continuing with platform.sh deployment...")
        else:
            # Quit and invite the user to try another platform.
            # We are happily exiting the script; there's no need to raise a
            #   CommandError.
            self.stdout.write(plsh_msgs.cancel_plsh)
            sys.exit()


    def validate_platform(self):
        """Make sure the local environment and project supports deployment to
        Platform.sh.
        
        The returncode for a successful command is 0, so anything truthy means
          a command errored out.
        """
        self._validate_cli()
        self._validate_platformshconfig()

        # When running unit tests, will not be logged into CLI.
        if not self.sd.unit_testing:
            self.deployed_project_name = self._get_platformsh_project_name()
            self.org_name = self._get_org_name()
        else:
            self.deployed_project_name = self.sd.deployed_project_name


    def prep_automate_all(self):
        """Do intial work for automating entire process.
        We know from validate_project() that user is logged into CLI.
        
        Returns:
        - None if creation was successful.
        - Raises CommandError if create command fails.

        Note: create command outputs project id to stdout if known, all other
          output goes to stderr.
        """

        self.sd.write_output("  Running `platform create`...")
        self.sd.write_output("    (Please be patient, this can take a few minutes.")
        cmd = f'platform create --title { self.deployed_project_name } --org {self.org_name} --region {self.sd.region} --yes'

        try:
            # Include check=True in this call, so we can process more
            #   significant errors that can come out of this call.
            # For example, if user can't create a project the 
            #   returncode will be 6, not 1. That causes subprocess.run()
            #   to actually raise an exception, not just pass a CompletedProcess
            #   instance.
            output = self.sd.execute_subp_run(cmd, check=True)
        except subprocess.CalledProcessError as e:
            error_msg = plsh_msgs.unknown_create_error(e)
            raise CommandError(error_msg)

        self.sd.write_output(output)


    # --- Helper methods for methods called from simple_deploy.py ---

    def _validate_cli(self):
        """Make sure the Platform.sh CLI is installed."""
        cmd = 'platform --version'
        output_obj = self.sd.execute_subp_run(cmd)
        if output_obj.returncode:
            raise CommandError(plsh_msgs.cli_not_installed)


    def _validate_platformshconfig(self):
        """If not using automate-all, make sure platformshconfig is installed
        locally.
        """
        if not self.sd.automate_all:
            cmd = 'pip show platformshconfig'
            output_obj = self.sd.execute_subp_run(cmd)
            if output_obj.returncode:
                raise CommandError(plsh_msgs.platformshconfig_not_installed)


    def _get_platformsh_project_name(self):
        """Get the deployed project name.
        This is the name that was given in the `platform create` command.
        - Try to get this from `project:info`.
        - If can't get project name:
          - Exit with warning, and inform user of --deployed-project-name
            flag to override this error.
        """
        # If we're creating the project, we'll just use the startprojet name.
        if self.sd.automate_all:
            return self.sd.project_name

        # Use the provided name if --deployed-project-name specified.
        if self.sd.deployed_project_name:
            return self.sd.deployed_project_name

        # Use --yes flag to avoid interactive prompt hanging in background
        #   if the user is not currently logged in to the CLI.
        cmd = "platform project:info --yes"
        output_obj = self.sd.execute_subp_run(cmd)
        output_str = output_obj.stdout.decode()

        # If there's no stdout, the user is probably logged out, hasn't called
        #   create, or doesn't have the CLI installed.
        # Also, I've seen both ProjectNotFoundException and RootNotFoundException
        #   raised when no project has been created.
        if not output_str:
            output_str = output_obj.stderr.decode()
            if 'LoginRequiredException' in output_str:
                raise CommandError(plsh_msgs.login_required)
            elif 'ProjectNotFoundException' in output_str:
                raise CommandError(plsh_msgs.no_project_name)
            elif 'RootNotFoundException' in output_str:
                raise CommandError(plsh_msgs.no_project_name)
            else:
                error_msg = plsh_msgs.unknown_error
                error_msg += plsh_msgs.cli_not_installed
                raise CommandError(error_msg)

        # Pull deployed project name from output.
        deployed_project_name_re = r'(\| title\s+?\|\s*?)(.*?)(\s*?\|)'
        match = re.search(deployed_project_name_re, output_str)
        if match:
            return match.group(2).strip()
        
        # Couldn't find a project name. Warn user, and let them know
        #   about override flag.
        raise CommandError(plsh_msgs.no_project_name)


    def _get_org_name(self):
        """Get the organization name associated with the user's Platform.sh
        account. This is needed for creating a project using automate_all.

        Confirm that it's okay to use this org.

        Returns:
        - None if not using automate-all.
        - String containing org name if found, and confirmed.
        - Raises CommandError if org name found, but not confirmed.
        - Raises CommandError with msg if CLI login required.
        - Raises CommandError with msg if org name not found.
        """
        if not self.sd.automate_all:
            return

        # Use --yes to suppress hanging at login prompt.
        cmd = "platform organization:list --yes"
        output_obj = self.sd.execute_subp_run(cmd)
        output_str = output_obj.stdout.decode()

        if not output_str:
            output_str = output_obj.stderr.decode()
            if 'LoginRequiredException' in output_str:
                raise CommandError(plsh_msgs.login_required)
            else:
                error_msg = plsh_msgs.unknown_error
                error_msg += plsh_msgs.cli_not_installed
                raise CommandError(error_msg)

        # Pull org name from output. Start by removing line containing lables.
        output_str_lines = [line for line in output_str.split("\n") if "Owner email" not in line]
        modified_output_str = '\n'.join(output_str_lines)
        org_name_re = r'(\|\s*)([a-zA-Z_]*)(.*)'
        match = re.search(org_name_re, modified_output_str)
        if match:
            org_name = match.group(2).strip()
            if self._confirm_use_org_name(org_name):
                return org_name
        else:
            # Got stdout, but can't find org id. Unknown error.
            raise CommandError(plsh_msgs.org_not_found)


    def _confirm_use_org_name(self, org_name):
        """Confirm that it's okay to use the org name that was found.
        Returns:
        - True if confirmed.
        - sys.exit() if not confirmed.
        """

        self.stdout.write(plsh_msgs.confirm_use_org_name(org_name))
        confirmed = self.sd.get_confirmation(skip_logging=True)

        if confirmed:
            self.stdout.write("  Okay, continuing with deployment.")
            return True
        else:
            # Exit, with a message that configuration is still an option.
            msg = plsh_msgs.cancel_plsh
            msg += plsh_msgs.may_configure
            raise CommandError(msg)



"""A collection of messages used in flyio.py."""

# For conventions, see documentation in deploy_messages.py

from textwrap import dedent

from django.conf import settings


confirm_preliminary = """
***** Deployments to Fly.io are experimental at this point ***

- Support for deploying to Fly.io is in an exploratory phase at this point.
- You should only be using this project to deploy to Fly.io at this point if
  you are interested in helping to develop or test the simple_deploy project.
- You should look at the deploy_flyio.py script before running this command,
  so you know what kinds of changes will be made to your project.
- This command will likely fail if you run it more than once.
- This command may not work if you already have a project deployed to Fly.io.
- You should understand the Fly.io console, and be comfortable deleting resources
  that are created during this deployment.
- You may want to cancel this run and deploy to a different platform.
"""

confirm_automate_all = """
The --automate-all flag means simple_deploy will:
- Run `fly apps create` for you, to create an empty Fly.io project.
- Configure your project for deployment on Fly.io.
- Commit all changes to your project that are necessary for deployment.
- Push these changes to Fly.io.
- Open your deployed project in a new browser tab.
"""

cancel_flyio = """
Okay, cancelling Fly.io deployment.
"""

# DEV: Update URL
# DEV: This could be moved to deploy_messages, with an arg for platform and URL.
cli_not_installed = """
In order to deploy to Fly.io, you need to install the Fly.io CLI.
  See here: fly_io_url
After installing the CLI, you can run simple_deploy again.
"""

no_project_name = """
A Fly.io app name could not be found.

The simple_deploy command expects that you've already created an app on Fly.io
to push to.

If you haven't done so, run the following command to create a new Fly.io app:

    $ flyctl apps create --generate-name

Then run simple_deploy again.
"""

create_app_failed = """
Could not create a Fly.io app.

The simple_deploy command can not proceed without a Fly.io app to deploy to.
You may have better luck with a configuration-only run, if you can create a Fly.io
app on your own. You may try the following command, and see if you can use any 
error messages to troubleshoot app creation:

    $ fly apps create --generate-name

If you can get this command to work, you can run simple_deploy again and the 
rest of the process may work.
"""

cancel_no_db = """
A database is required for deployment. You may be able to create a database
manually, and configure it to work with this app.
"""


# --- Dynamic strings ---
# These need to be generated in functions, to display information that's 
#   determined as the script runs.

def region_not_found(app_name):
    """Could not find a region to deploy to."""

    msg = dedent(f"""
        --- A Fly.io region was not found. ---

        We need to know what region the app is going to be deployed to.
        We could not find a region in the output of:

        $ flyctl regions list -a {app_name}
    """)

    return msg


def confirm_use_org_name(org_name):
    """Confirm use of this org name to create a new project."""

    msg = dedent(f"""
        --- The Platform.sh CLI requires an organization name when creating a new project. ---
        When using --automate-all, a project will be created on your behalf. The following
        organization name was found: {org_name}

        This organization will be used to create a new project. If this is not okay,
        enter n to cancel this operation.
    """)

    return msg


def confirm_create_db(db_cmd):
    """Confirm it's okay to create a Postgres database on the user's account."""

    msg = dedent(f"""
        A Postgres database is required to continue with deployment. If you confirm this,
        the following command will be run, to create a new database on your account:
        $ {db_cmd}
    """)

    return msg


def success_msg(log_output=''):
    """Success message, for configuration-only run."""

    msg = dedent(f"""
        --- Your project is now configured for deployment on Fly.io ---

        To deploy your project, you will need to:
        - Commit the changes made in the configuration process.
            $ git status
            $ git add .
            $ git commit -am "Configured project for deployment."
        - Push your project to Fly.io's servers:
            $ fly deploy
        - Open your project:
            $ fly open    
        - As you develop your project further:
            - Make local changes
            - Commmit your local changes
            - Run `fly deploy`
    """)

    if log_output:
        msg += dedent(f"""
        - You can find a full record of this configuration in the simple_deploy_logs directory.
        """)

    return msg


def success_msg_automate_all(deployed_url):
    """Success message, when using --automate-all."""

    msg = dedent(f"""

        --- Your project should now be deployed on Fly.io ---

        It should have opened up in a new browser tab.
        - You can also visit your project at {deployed_url}

        If you make further changes and want to push them to Fly.io,
        commit your changes and then run `fly deploy`.
    """)
    return msg




"""Manages all Fly.io-specific aspects of the deployment process."""

# Note: Internal references to Fly.io will almost always be flyio.
#   Public references, such as the --platform argument, will be fly_io.

import sys, os, re, subprocess
from pathlib import Path

from django.conf import settings
from django.core.management.base import CommandError
from django.core.management.utils import get_random_secret_key
from django.utils.crypto import get_random_string
from django.template.engine import Engine
from django.template.loaders.app_directories import Loader
from django.template.loader import render_to_string
from django.utils.safestring import mark_safe

from simple_deploy.management.commands.utils import deploy_messages as d_msgs
from simple_deploy.management.commands.utils import deploy_messages_flyio as flyio_msgs


class FlyioDeployer:
    """Perform the initial deployment of a simple project.
    Configure as much as possible automatically.
    """

    def __init__(self, command):
        """Establishes connection to existing simple_deploy command object."""
        self.sd = command
        self.stdout = self.sd.stdout


    def deploy(self, *args, **options):
        self.sd.write_output("Configuring project for deployment to Fly.io...")

        self._set_on_flyio()
        self._set_debug()

        self._add_dockerfile()
        self._add_flytoml_file()
        self._modify_settings()

        self._add_gunicorn()
        self._add_psycopg2()
        self._add_dj_database_url()

        self._conclude_automate_all()

        self._show_success_message()


    def _set_on_flyio(self):
        """Set a secret, ON_FLYIO. This is used in settings.py to apply
        deployment-specific settings.
        Returns:
        - None
        """
        msg = "Setting ON_FLYIO secret..."
        self.sd.write_output(msg)

        # First check if secret has already been set.
        cmd = f"flyctl secrets list -a {self.deployed_project_name}"
        output_obj = self.sd.execute_subp_run(cmd)
        output_str = output_obj.stdout.decode()
        if 'ON_FLYIO' in output_str:
            msg = "  Found ON_FLYIO in existing secrets."
            self.sd.write_output(msg)
            return

        cmd = f"flyctl secrets set -a {self.deployed_project_name} ON_FLYIO=1"
        output_obj = self.sd.execute_subp_run(cmd)
        output_str = output_obj.stdout.decode()
        self.sd.write_output(output_str)

        msg = "  Set ON_FLYIO secret."
        self.sd.write_output(msg)


    def _set_debug(self):
        """Set a secret, DEBUG=FALSE. This is used in settings.py to apply
        deployment-specific settings.
        Returns:
        - None
        """
        msg = "Setting DEBUG secret..."
        self.sd.write_output(msg)

        # First check if secret has already been set.
        cmd = f"flyctl secrets list -a {self.deployed_project_name}"
        output_obj = self.sd.execute_subp_run(cmd)
        output_str = output_obj.stdout.decode()
        if 'DEBUG' in output_str:
            msg = "  Found DEBUG in existing secrets."
            self.sd.write_output(msg)
            return

        cmd = f"flyctl secrets set -a {self.deployed_project_name} DEBUG=FALSE"
        output_obj = self.sd.execute_subp_run(cmd)
        output_str = output_obj.stdout.decode()
        self.sd.write_output(output_str)

        msg = "  Set DEBUG=FALSE secret."
        self.sd.write_output(msg)


    def _add_dockerfile(self):
        """Add a minimal dockerfile."""

        # File should be in project root, if present.
        self.sd.write_output(f"\n  Looking in {self.sd.git_path} for Dockerfile...")
        dockerfile_present = 'Dockerfile' in os.listdir(self.sd.git_path)

        if dockerfile_present:
            self.sd.write_output("    Found existing Dockerfile.")
        else:
            # Generate file from template.
            self.sd.write_output("    No Dockerfile found. Generating file...")
            my_loader = Loader(Engine.get_default())
            my_template = my_loader.get_template('dockerfile_flyio')

            # Build context dict for template.
            context = {
                'django_project_name': self.sd.project_name, 
                }
            template_string = render_to_string('dockerfile_flyio', context)

            path = self.sd.project_root / 'Dockerfile'
            path.write_text(template_string)

            msg = f"\n    Generated Dockerfile: {path}"
            self.sd.write_output(msg)
            return path


    def _add_flytoml_file(self):
        """Add a minimal fly.toml file."""
        # File should be in project root, if present.
        self.sd.write_output(f"\n  Looking in {self.sd.git_path} for fly.toml file...")
        flytoml_present = 'fly.toml' in os.listdir(self.sd.git_path)

        if flytoml_present:
            self.sd.write_output("    Found existing fly.toml file.")
        else:
            # Generate file from template.
            self.sd.write_output("    No fly.toml file found. Generating file...")
            my_loader = Loader(Engine.get_default())
            my_template = my_loader.get_template('fly.toml')

            # Build context dict for template.
            context = {
                'deployed_project_name': self.deployed_project_name, 
                }
            template_string = render_to_string('fly.toml', context)

            path = self.sd.project_root / 'fly.toml'
            path.write_text(template_string)

            msg = f"\n    Generated fly.toml: {path}"
            self.sd.write_output(msg)
            return path


    def _modify_settings(self):
        """Add settings specific to Fly.io."""
        #   Check if a fly.io section is present. If not, add settings. If already present,
        #   do nothing.
        self.sd.write_output("\n  Checking if settings block for Fly.io present in settings.py...")

        with open(self.sd.settings_path) as f:
            settings_string = f.read()

        if 'if os.environ.get("ON_FLYIO"):' in settings_string:
            self.sd.write_output("\n    Found Fly.io settings block in settings.py.")
            return

        # Add Fly.io settings block.
        self.sd.write_output("    No Fly.io settings found in settings.py; adding settings...")
        my_loader = Loader(Engine.get_default())
        my_template = my_loader.get_template('flyio_settings.py')

        # Build context dict for template.
        safe_settings_string = mark_safe(settings_string)
        context = {
            'current_settings': safe_settings_string,
            'deployed_project_name': self.deployed_project_name,
        }
        template_string = render_to_string('flyio_settings.py', context)

        path = Path(self.sd.settings_path)
        path.write_text(template_string)

        msg = f"    Modified settings.py file: {path}"
        self.sd.write_output(msg)


    def _add_gunicorn(self):
        """Add gunicorn to project requirements."""
        self.sd.write_output("\n  Looking for gunicorn...")

        if self.sd.using_req_txt:
            self.sd.add_req_txt_pkg('gunicorn')
        elif self.sd.using_pipenv:
            self.sd.add_pipenv_pkg('gunicorn')


    def _add_psycopg2(self):
        """Add psycopg2 to project requirements."""
        self.sd.write_output("\n  Looking for psycopg2...")

        if self.sd.using_req_txt:
            self.sd.add_req_txt_pkg('psycopg2')
        elif self.sd.using_pipenv:
            self.sd.add_pipenv_pkg('psycopg2')

    def _add_dj_database_url(self):
        """Add dj-database-url to project requirements."""
        self.sd.write_output("\n  Looking for dj-database-url...")

        if self.sd.using_req_txt:
            self.sd.add_req_txt_pkg('dj-database-url')
        elif self.sd.using_pipenv:
            self.sd.add_pipenv_pkg('dj-database-url')


    def _conclude_automate_all(self):
        """Finish automating the push to Fly.io.
        - Commit all changes.
        - Call `fly deploy`.
        - Call `fly open`, and grab URL.
        """
        # Making this check here lets deploy() be cleaner.
        if not self.sd.automate_all:
            return

        self.sd.commit_changes()

        # Push project.
        self.sd.write_output("  Deploying to Fly.io...")
        cmd = "fly deploy"
        output = self.sd.execute_subp_run(cmd)
        self.sd.write_output(output)

        # Open project.
        self.sd.write_output("  Opening deployed app in a new browser tab...")
        cmd = "fly open"
        output = self.sd.execute_subp_run(cmd)
        self.sd.write_output(output)

        # Get url of deployed project.
        url_re = r'(opening )(http.*?)( \.\.\.)'
        output_str = output.stdout.decode()
        m = re.search(url_re, output_str)
        if m:
            self.deployed_url = m.group(2).strip()


    def _show_success_message(self):
        """After a successful run, show a message about what to do next."""

        # DEV:
        # - Mention that this script should not need to be run again, unless
        #   creating a new deployment.
        #   - Describe ongoing approach of commit, push, migrate. Lots to consider
        #     when doing this on production app with users, make sure you learn.

        if self.sd.automate_all:
            msg = flyio_msgs.success_msg_automate_all(self.deployed_url)
            self.sd.write_output(msg)
        else:
            msg = flyio_msgs.success_msg(log_output=self.sd.log_output)
            self.sd.write_output(msg)


    # --- Methods called from simple_deploy.py ---

    def confirm_preliminary(self):
        """Deployment to Fly.io is in a preliminary state, and we need to be
        explicit about that.
        """
        # Skip this confirmation when unit testing.
        if self.sd.unit_testing:
            return

        self.stdout.write(flyio_msgs.confirm_preliminary)
        confirmed = self.sd.get_confirmation(skip_logging=True)

        if confirmed:
            self.stdout.write("  Continuing with Fly.io deployment...")
        else:
            # Quit and invite the user to try another platform.
            # We are happily exiting the script; there's no need to raise a
            #   CommandError.
            self.stdout.write(flyio_msgs.cancel_flyio)
            sys.exit()


    def validate_platform(self):
        """Make sure the local environment and project supports deployment to
        Fly.io.
        
        The returncode for a successful command is 0, so anything truthy means
          a command errored out.
        """
        self._validate_cli()

        # When running unit tests, will not be logged into CLI.
        if not self.sd.unit_testing:
            self.deployed_project_name = self._get_deployed_project_name()

            # If using automate_all, we need to create the app before creating
            #   the db. But if there's already an app with no deployment, we can 
            #   use that one (maybe created from a previous automate_all run).
            # DEV: Update _get_deployed_project_name() to not throw error if
            #   using automate_all. _create_flyio_app() can exit if not using
            #   automate_all(). If self.deployed_project_name is set, just return
            #   because we'll use that project. If it's not set, call create.
            if not self.deployed_project_name and self.sd.automate_all:
                self.deployed_project_name = self._create_flyio_app()

            # Create the db now, before any additional configuration. Get region
            #   so we know where to create the db.
            self.region = self._get_region()
            self._create_db()
        else:
            self.deployed_project_name = self.sd.deployed_project_name


    def prep_automate_all(self):
        """Take any further actions needed if using automate_all."""
        # All creation has been taken earlier, during validation.
        pass


    # --- Helper methods for methods called from simple_deploy.py ---

    def _validate_cli(self):
        """Make sure the Platform.sh CLI is installed."""
        cmd = 'flyctl version'
        output_obj = self.sd.execute_subp_run(cmd)
        if output_obj.returncode:
            raise CommandError(flyio_msgs.cli_not_installed)

    def _get_deployed_project_name(self):
        """Get the Fly.io project name.
        Parse the output of `flyctl apps list`, and look for an app name
          that doesn't have a value set for LATEST DEPLOY. This indicates
          an app that has just been created, and has not yet been deployed.

        Returns:
        - String representing deployed project name.
        - Empty string if no deployed project name found, but using automate_all.
        - Raises CommandError if deployed project name can't be found.
        """
        msg = "\nLooking for Fly.io app to deploy against..."
        self.sd.write_output(msg, skip_logging=True)

        # Get apps info.
        cmd = "flyctl apps list"
        output_obj = self.sd.execute_subp_run(cmd)
        output_str = output_obj.stdout.decode()

        # Only keep relevant output; get rid of blank lines, update messages,
        #   and line with labels like NAME and LATEST DEPLOY.
        lines = output_str.split('\n')
        lines = [line for line in lines if line]
        lines = [line for line in lines if 'update' not in line.lower()]
        lines = [line for line in lines if 'NAME' not in line]
        lines = [line for line in lines if 'builder' not in line]

        # An app that has not been deployed to will only have values set for NAME,
        #   OWNER, and STATUS. PLATFORM and LATEST DEPLOY will be empty.
        app_name = ''
        for line in lines:
            # The desired line has three elements.
            parts = line.split()
            if len(parts) == 3:
                app_name = parts[0]

        # Return deployed app name, or raise CommandError.
        if app_name:
            msg = f"  Found Fly.io app: {app_name}"
            self.sd.write_output(msg, skip_logging=True)
            return app_name
        elif self.sd.automate_all:
            msg = "  No app found, but continuing with --automate-all..."
            self.sd.write_output(msg, skip_logging=True)
            # Simply return an empty string to indicate no suitable app was found,
            #   and we'll create one later.
            return ""
        else:
            # Can't continue without a Fly.io app to configure against.
            raise CommandError(flyio_msgs.no_project_name)

    def _create_flyio_app(self):
        """Create a new Fly.io app.
        Assumes caller already checked for automate_all, and that a suitable
          app is not already available.
        Returns:
        - String representing new app name.
        - Raises CommandError if an app can't be created.
        """
        msg = "  Creating a new app on Fly.io..."
        self.sd.write_output(msg, skip_logging=True)

        cmd = "flyctl apps create --generate-name"
        output_obj = self.sd.execute_subp_run(cmd)
        output_str = output_obj.stdout.decode()
        self.sd.write_output(output_str, skip_logging=True)

        # Get app name.
        app_name_re = r'(New app created: )(\w+\-\w+\-\d+)'
        flyio_app_name = ''
        m = re.search(app_name_re, output_str)
        if m:
            flyio_app_name = m.group(2).strip()

        if flyio_app_name:
            msg = f"  Created new app: {flyio_app_name}"
            self.sd.write_output(msg, skip_logging=True)
            return flyio_app_name
        else:
            # Can't continue without a Fly.io app to deploy to.
            raise CommandError(flyio_msgs.create_app_failed)


    def _get_region(self):
        """Get the region that the Fly.io app is configured for. We'll need this
        to create a postgres database.

        Parse the output of `flyctl regions list -a app_name`.

        Returns:
        - String representing region.
        - Raises CommandError if can't find region.
        """

        msg = "Looking for Fly.io region..."
        self.sd.write_output(msg, skip_logging=True)

        # Get region output.
        cmd = f"flyctl regions list -a {self.deployed_project_name}"
        output_obj = self.sd.execute_subp_run(cmd)
        output_str = output_obj.stdout.decode()

        # Look for first three-letter line after Region Pool.
        region = ''
        pool_found = False
        lines = output_str.split('\n')
        for line in lines:
            if not pool_found and "Region Pool" in line:
                pool_found = True
                continue

            # This is the first line after Region Pool.
            if pool_found:
                region = line.strip()
                break

        # Return region name, or raise CommandError.
        if region:
            msg = f"  Found region: {region}"
            self.sd.write_output(msg, skip_logging=True)
            return region
        else:
            # Can't continue without a Fly.io region to configure against.
            raise CommandError(flyio_msgs.region_not_found(self.deployed_project_name))

    def _create_db(self):
        """Create a db to deploy to, if none exists.
        Returns: 
        - 
        - Raises CommandError if...
        """
        msg = "Looking for a Postgres database..."
        self.sd.write_output(msg, skip_logging=True)

        db_exists = self._check_if_db_exists()

        if db_exists:
            return

        # No db found, create a new db.
        msg = f"  Create a new Postgres database..."
        self.sd.write_output(msg, skip_logging=True)

        self.db_name = f"{self.deployed_project_name}-db"
        cmd = f"flyctl postgres create --name {self.db_name} --region {self.region}"
        cmd += " --initial-cluster-size 1 --vm-size shared-cpu-1x --volume-size 1"

        # If not using automate_all, make sure it's okay to create a resource
        #   on user's account.
        if not self.sd.automate_all:
            self._confirm_create_db(db_cmd=cmd)

        # Create database and parse output.
        output_obj = self.sd.execute_subp_run(cmd)
        output_str = output_obj.stdout.decode()
        self.sd.write_output(output_str, skip_logging=True)

        msg = "  Created Postgres database."
        self.sd.write_output(msg, skip_logging=True)

        # Run `attach` command (and confirm DATABASE_URL is set?)
        msg = "  Attaching database to Fly.io app..."
        self.sd.write_output(msg, skip_logging=True)
        cmd = f"flyctl postgres attach --app {self.deployed_project_name} {self.db_name}"

        output_obj = self.sd.execute_subp_run(cmd)
        output_str = output_obj.stdout.decode()
        self.sd.write_output(output_str, skip_logging=True)

        msg = "  Attached database to app."
        self.sd.write_output(msg, skip_logging=True)

    def _check_if_db_exists(self):
        """Check if a postgres db already exists that should be used with this app.
        Returns:
        - True if db found.
        - False if not found.
        """

        # First, see if any Postgres clusters exist.
        cmd = "flyctl postgres list"
        output_obj = self.sd.execute_subp_run(cmd)
        output_str = output_obj.stdout.decode()

        if "No postgres clusters found" in output_str:
            msg = "  No Postgres database found."
            self.sd.write_output(msg, skip_logging=True)
            return False
        else:
            msg = "  A Postgres database was found."
            self.sd.write_output(msg, skip_logging=True)
            return True

    def _confirm_create_db(self, db_cmd):
        """We really need to confirm that the user wants a db created on their behalf.
        Show the command that will be run on the user's behalf.
        Returns:
        - True if confirmed.
        - Raises CommandError if not confirmed.
        """
        if self.sd.unit_testing:
            return

        self.stdout.write(flyio_msgs.confirm_create_db(db_cmd))
        confirmed = self.sd.get_confirmation(skip_logging=True)

        if confirmed:
            self.stdout.write("  Creating database...")
        else:
            # Quit and invite the user to create a database manually.
            raise CommandError(flyio_msgs.cancel_no_db)


"""A collection of messages used in deploy_platformsh.py."""

# For conventions, see documentation in deploy_messages.py

from textwrap import dedent

from django.conf import settings


confirm_preliminary = """
***** Deployments to platform.sh are experimental at this point ***

- Support for deploying to platform.sh is in an exploratory phase at this point.
- You should only be using this project to deploy to platform.sh at this point if
  you are interested in helping to develop or test the simple_deploy project.
- You should look at the deploy_platformsh.py script before running this command,
  so you know what kinds of changes will be made to your project.
- You should understand the platform.sh console, and be comfortable deleting resources
  that are created during this deployment.
- You may want to cancel this run and deploy to a different platform.
"""

confirm_automate_all = """
The --automate-all flag means simple_deploy will:
- Run `platform create` for you, to create an empty Platform.sh project.
  - This will create a project in the us-3.platform.sh region. If you wish
    to use a different region, cancel this operation and use the --region flag.
  - You can see a list of all regions by running `platform help project:create`
- Commit all changes to your project that are necessary for deployment.
- Push these changes to Platform.sh.
- Open your deployed project in a new browser tab.
"""

cancel_plsh = """
Okay, cancelling platform.sh deployment.
"""

cli_not_installed = """
In order to deploy to Platform.sh, you need to install the Platform.sh CLI.
  See here: https://docs.platform.sh/gettingstarted/introduction/template/cli-requirements.html
After installing the CLI, you can run simple_deploy again.
"""

platformshconfig_not_installed = """
In order to deploy to Platform.sh, you need to install the package `platformshoconfig`.
To do this, run the following command, or its equivalent in your project:
    pip install platformshconfig
After installing this package, you can run simple_deploy again.
"""

no_project_name = """
A Platform.sh project name could not be found.

The simple_deploy command expects that you've already run `platform create`, or
associated the local project with an existing project on Platform.sh.

If you haven't done so, run the `platform create` command and then run
simple_deploy again. You can override this warning by using
the `--deployed-project-name` flag to specify the name you want to use for the
project. This must match the name of your Platform.sh project.
"""

org_not_found = """
A Platform.sh organization name could not be found.

You may have created a Platform.sh account, but not created an organization.
The Platform.sh CLI requires an organization name when creating a new project.

Please visit the Platform.sh console and make sure you have created an organization.
You can also do this through the CLI using the `platform organization:create` command.
For help, run `platform help organization:create`.
"""

login_required = """
You appear to be logged out of the Platform.sh CLI. Please run the 
command `platform login`, and then run simple_deploy again.

You may be able to override this error by passing the `--deployed-project-name`
flag.
"""

unknown_error = """
An unknown error has occurred. Do you have the Platform.sh CLI installed?
"""

may_configure = """
You may want to re-run simple_deploy without the --automate-all flag.

You will have to create the Platform.sh project yourself, but simple_deploy
will do all of the necessary configuration for deployment.
"""


# --- Dynamic strings ---
# These need to be generated in functions, to display information that's 
#   determined as the script runs.

def confirm_use_org_name(org_name):
    """Confirm use of this org name to create a new project."""

    msg = dedent(f"""
        --- The Platform.sh CLI requires an organization name when creating a new project. ---
        When using --automate-all, a project will be created on your behalf. The following
        organization name was found: {org_name}

        This organization will be used to create a new project. If this is not okay,
        enter n to cancel this operation.
    """)

    return msg


def unknown_create_error(e):
    """Process a non-specific error when running `platform create` 
    while using automate_all. This is most likely an issue with the user
    not having permission to create a new project, for example because they
    are on a trial plan and have already created too many projects.
    """

    msg = dedent(f"""
        --- An error has occurred when trying to create a new Platform.sh project. ---

        While running `platform create`, an error has occurred. You should check
        the Platform.sh console to see if a project was partially created.

        The error messages that Platform.sh provides, both through the CLI and
        the console, are not always specific enough to be helpful. For example, 
        newer users are limited to two new projects in a 24-hour period, or something
        like that. But if you try to create an additional project, you only get
        a message that says: "You do not have access to create a new Subscriptions resource".
        There is no information about specific limits, and how to address them.

        The following output may help diagnose the error:
        ***** output of `platform create` *****

        {e.stderr.decode()}

        ***** end output *****
    """)

    return msg


def success_msg(log_output=''):
    """Success message, for configuration-only run."""

    msg = dedent(f"""
        --- Your project is now configured for deployment on Platform.sh. ---

        To deploy your project, you will need to:
        - Commit the changes made in the configuration process.
            $ git status
            $ git add .
            $ git commit -am "Configured project for deployment."
        - Push your project to Platform.sh' servers:
            $ platform push
        - Open your project:
            $ platform url    
        - As you develop your project further:
            - Make local changes
            - Commmit your local changes
            - Run `platform push`
    """)

    if log_output:
        msg += dedent(f"""
        - You can find a full record of this configuration in the simple_deploy_logs directory.
        """)

    return msg


def success_msg_automate_all(deployed_url):
    """Success message, when using --automate-all."""

    msg = dedent(f"""

        --- Your project should now be deployed on Platform.sh. ---

        It should have opened up in a new browser tab.
        - You can also visit your project at {deployed_url}

        If you make further changes and want to push them to Platform.sh,
        commit your changes and then run `platform push`.

        Also, if you haven't already done so you should review the
        documentation for Python deployments on Platform.sh at:
        - https://docs.platform.sh/languages/python.html
        - This documentation will help you understand how to maintain
          your deployment.

    """)
    return msg


"""A collection of messages used in simple_deploy.py."""

# Module is called deploy_messages to avoid collisions with any messages.py 
#   files.
#
# Storing messages here makes for a much shorter simple_deploy.py file, and
#   makes it easier to see how the messages actually look on screen. The 
#   disadvantage is that it's harder to track what messages are being
#   displayed when working on simple_deploy.
# When you add a message here, make sure the stdout.write() call has a comment
#   that clearly summarizes what the message says.
#
# CommandError messages
# - Any message used when raising a CommandError should have a blank line after
#   the line with the opening triple quote, to separate the main error message
#   from the generic "Command Error".
#
# Line length conventions
# - Try to keep lines at the 79-char PEP 8 limit, but the project is not
#   overly strict about it at this point.
#
# - DynamicMessages
#   - It can be helpful to make a second vertical line at 86 characters
#     (78 characters + two indentation levels) to judge the line lengths
#     for dynamic messages.


from textwrap import dedent

from django.conf import settings


cancel_automate_all = """
Okay, canceling this run. If you want to configure your project
for deployment, run simple_deploy again without the --automate-all flag.
"""

requires_platform_flag = """
The --platform flag is required; you must specify which platform you want to 
deploy your project to.
- Current options are `heroku` and `platform_sh`.
- Example usage:
  $ python manage.py simple_deploy --platform heroku
  $ python manage.py simple_deploy --platform platform_sh

For more detailed information, see https://github.com/ehmatthes/django-simple-deploy/blob/main/docs/cli_args.md

Please re-run the command with a --platform option specified.
"""

unclean_git_status = """
The output of `git status` indicates that you have uncommitted changes.

We highly recommend you commit all changes before running simple_deploy, so
you can easily undo configuration changes if deployment doesn't work, or if you
wish to deploy to a different platform.

Please commit all changes and then run simple_deploy again. If you really wish
to continue without committing current changes, you can use
the --ignore-unclean-git flag.
"""

# Add-on to `unclean_git_status` when using automate-all.
unclean_git_automate_all = """
It's especially important to have a clean status when using the --automate-all
flag. In this case, simple_deploy will make commits on your behalf. It's a
really good idea to keep your commits separate from the commits that simple_deploy
makes.
"""


# --- Dynamic strings ---
# These need to be generated in functions, to display information that's 
#   determined as the script runs.

# No platform-agnostic dynamic strings are needed at this time.
#   However, some messages that are currently in platform-specific files
#   can probably be moved here.


"""Manages all Heroku-specific aspects of the deployment process."""

import sys, os, re, subprocess

from django.conf import settings
from django.core.management.base import CommandError
from django.core.management.utils import get_random_secret_key
from django.utils.crypto import get_random_string

from simple_deploy.management.commands.utils import deploy_messages as d_msgs
from simple_deploy.management.commands.utils import deploy_messages_heroku as dh_msgs


class HerokuDeployer:
    """Perform the initial deployment of a simple project.
    Configure as much as possible automatically.
    """

    def __init__(self, command):
        """Establishes connection to existing simple_deploy command object."""
        self.sd = command
        self.stdout = self.sd.stdout


    def deploy(self, *args, **options):
        self.sd.write_output("Configuring project for deployment to Heroku...")

        self._get_heroku_app_info()
        self._set_heroku_env_var()
        self._get_heroku_settings()
        self._generate_procfile()
        self._add_gunicorn()
        self._check_allowed_hosts()
        self._configure_db()
        self._configure_static_files()
        self._configure_debug()
        self._configure_secret_key()
        self._conclude_automate_all()
        self._summarize_deployment()
        self._show_success_message()


    # --- Methods used in this class ---

    def _get_heroku_app_info(self):
        """Get info about the Heroku app we're pushing to."""
        # We assume the user has already run 'heroku create', or --automate-all
        #   has run it. If it hasn't been run, we'll quit and tell them to do so.

        # DEV: The testing approach here should be improved. We should be able
        #   to easily test for a failed apps:info call. Also, probably want
        #   to mock the output of apps:info rather than directly setting
        #   heroku_app_name.
        if self.sd.unit_testing:
            self.heroku_app_name = 'sample-name-11894'
        else:
            self.sd.write_output("  Inspecting Heroku app...")
            cmd = 'heroku apps:info'
            apps_info = self.sd.execute_subp_run(cmd)
            self.sd.write_output(apps_info)

            # Turn stdout info into a list of strings that we can then parse.
            #   If no app exists, stdout is empty and the output went to stderr.
            apps_info = apps_info.stdout.decode().split('\n')
            # DEV: Use this code when we can require Python >=3.9.
            # self.heroku_app_name = apps_info[0].removeprefix('=== ')
            self.heroku_app_name = apps_info[0].replace('=== ', '')

        if self.heroku_app_name:
            self.sd.write_output(f"    Found Heroku app: {self.heroku_app_name}")
        else:
            # Let user know they need to run `heroku create`.
            self.sd.write_output(dh_msgs.no_heroku_app_detected,
                write_to_console=False)
            raise CommandError(dh_msgs.no_heroku_app_detected)


    def _set_heroku_env_var(self):
        """Set a config var to indicate when we're in the Heroku environment.
        This is mostly used to modify settings for the deployed project.
        """

        # Skip this entirely when unit testing.
        if self.sd.unit_testing:
            return

        self.sd.write_output("  Setting Heroku environment variable...")
        cmd = 'heroku config:set ON_HEROKU=1'
        output = self.sd.execute_subp_run(cmd)
        self.sd.write_output(output)
        self.sd.write_output("    Set ON_HEROKU=1.")
        self.sd.write_output("    This is used to define Heroku-specific settings.")


    def _get_heroku_settings(self):
        """Get any heroku-specific settings that are already in place.
        """
        # If any heroku settings have already been written, we don't want to
        #  add them again. This assumes a section at the end, starting with a
        #  check for 'ON_HEROKU' in os.environ.

        with open(self.sd.settings_path) as f:
            settings_lines = f.readlines()

        self.found_heroku_settings = False
        self.current_heroku_settings_lines = []
        for line in settings_lines:
            if "if 'ON_HEROKU' in os.environ:" in line:
                self.found_heroku_settings = True
            if self.found_heroku_settings:
                self.current_heroku_settings_lines.append(line)


    def _generate_procfile(self):
        """Create Procfile, if none present."""

        #   Procfile should be in project root, if present.
        self.sd.write_output(f"\n  Looking in {self.sd.git_path} for Procfile...")
        procfile_present = 'Procfile' in os.listdir(self.sd.git_path)

        if procfile_present:
            self.sd.write_output("    Found existing Procfile.")
        else:
            self.sd.write_output("    No Procfile found. Generating Procfile...")
            if self.sd.nested_project:
                proc_command = f"web: gunicorn {self.sd.project_name}.{self.sd.project_name}.wsgi --log-file -"
            else:
                proc_command = f"web: gunicorn {self.sd.project_name}.wsgi --log-file -"

            with open(f"{self.sd.git_path}/Procfile", 'w') as f:
                f.write(proc_command)

            self.sd.write_output("    Generated Procfile with following process:")
            self.sd.write_output(f"      {proc_command}")


    def _add_gunicorn(self):
        """Add gunicorn to project requirements."""
        self.sd.write_output("\n  Looking for gunicorn...")

        if self.sd.using_req_txt:
            self.sd.add_req_txt_pkg('gunicorn')
        elif self.sd.using_pipenv:
            self.sd.add_pipenv_pkg('gunicorn')


    def _check_allowed_hosts(self):
        """Make sure project can be served from heroku."""
        # This method is specific to Heroku.

        self.sd.write_output("\n  Making sure project can be served from Heroku...")
        heroku_host = f"{self.heroku_app_name}.herokuapp.com"

        if heroku_host in settings.ALLOWED_HOSTS:
            self.sd.write_output(f"    Found {heroku_host} in ALLOWED_HOSTS.")
        elif 'herokuapp.com' in settings.ALLOWED_HOSTS:
            # This is a generic entry that allows serving from any heroku URL.
            self.sd.write_output("    Found 'herokuapp.com' in ALLOWED_HOSTS.")
        else:
            new_setting = f"ALLOWED_HOSTS.append('{heroku_host}')"
            msg_added = f"    Added {heroku_host} to ALLOWED_HOSTS for the deployed project."
            msg_already_set = f"    Found {heroku_host} in ALLOWED_HOSTS for the deployed project."
            self._add_heroku_setting(new_setting, msg_added, msg_already_set)


    def _configure_db(self):
        """Add required db-related packages, and modify settings for Heroku db.
        """
        self.sd.write_output("\n  Configuring project for Heroku database...")
        self._add_db_packages()
        self._add_db_settings()


    def _add_db_packages(self):
        """Add packages required for the Heroku db."""
        self.sd.write_output("    Adding db-related packages...")

        # psycopg2 2.9 causes "database connection isn't set to UTC" issue.
        #   See: https://github.com/ehmatthes/heroku-buildpack-python/issues/31
        if self.sd.using_req_txt:
            self.sd.add_req_txt_pkg('psycopg2<2.9')
            self.sd.add_req_txt_pkg('dj-database-url')
        elif self.sd.using_pipenv:
            self.sd.add_pipenv_pkg('psycopg2', version="<2.9")
            self.sd.add_pipenv_pkg('dj-database-url')


    def _add_db_settings(self):
        """Add settings for Heroku db."""
        self.sd.write_output("   Checking Heroku db settings...")

        # Import dj-database-url.
        new_setting = "import dj_database_url"
        msg_added = "    Added import statement for dj-database-url."
        msg_already_set = "    Found import statement for dj-database-url."
        self._add_heroku_setting(new_setting, msg_added, msg_already_set)

        # Configure db.
        new_setting = "DATABASES = {'default': dj_database_url.config(default='postgres://localhost')}"
        msg_added = "    Added setting to configure Postgres on Heroku."
        msg_already_set = "    Found setting to configure Postgres on Heroku."
        self._add_heroku_setting(new_setting, msg_added, msg_already_set)


    def _configure_static_files(self):
        """Configure static files for Heroku deployment."""

        self.sd.write_output("\n  Configuring static files for Heroku deployment...")

        # Add whitenoise to requirements.
        self.sd.write_output("    Adding staticfiles-related packages...")
        if self.sd.using_req_txt:
            self.sd.add_req_txt_pkg('whitenoise')
        elif self.sd.using_pipenv:
            self.sd.add_pipenv_pkg('whitenoise')

        # Modify settings, and add a directory for static files.
        self._add_static_file_settings()
        self._add_static_file_directory()


    def _add_static_file_settings(self):
        """Add all settings needed to manage static files."""
        self.sd.write_output("    Configuring static files settings...")

        new_setting = "STATIC_ROOT = os.path.join(BASE_DIR, 'staticfiles')"
        msg_added = "    Added STATIC_ROOT setting for Heroku."
        msg_already_set = "    Found STATIC_ROOT setting for Heroku."
        self._add_heroku_setting(new_setting, msg_added, msg_already_set)

        new_setting = "STATIC_URL = '/static/'"
        msg_added = "    Added STATIC_URL setting for Heroku."
        msg_already_set = "    Found STATIC_URL setting for Heroku."
        self._add_heroku_setting(new_setting, msg_added, msg_already_set)

        new_setting = "STATICFILES_DIRS = (os.path.join(BASE_DIR, 'static'),)"
        msg_added = "    Added STATICFILES_DIRS setting for Heroku."
        msg_already_set = "    Found STATICFILES_DIRS setting for Heroku."
        self._add_heroku_setting(new_setting, msg_added, msg_already_set)


    def _add_static_file_directory(self):
        """Create a folder for static files, if it doesn't already exist.
        """
        self.sd.write_output("    Checking for static files directory...")

        # Make sure there's a static files directory.
        static_files_dir = f"{self.sd.project_root}/static"
        if os.path.exists(static_files_dir):
            if os.listdir(static_files_dir):
                self.sd.write_output("    Found non-empty static files directory.")
                return
        else:
            os.makedirs(static_files_dir)
            self.sd.write_output("    Created empty static files directory.")

        # Add a placeholder file to the empty static files directory.
        placeholder_file = f"{static_files_dir}/placeholder.txt"
        with open(placeholder_file, 'w') as f:
            f.write("This is a placeholder file to make sure this folder is pushed to Heroku.")
        self.sd.write_output("    Added placeholder file to static files directory.")


    def _configure_debug(self):
        """Use an env var to manage DEBUG setting, and set to False."""

        # Config variables are strings, which always causes confusion for people
        #   when setting boolean env vars. A good habit is to use something other than
        #   True or False, so it's clear we're not trying to use Python's default
        #   boolean values.
        # Here we use 'TRUE' and 'FALSE'. Then a simple test:
        #    os.environ.get('DEBUG') == 'TRUE'
        # returns the bool value True for 'TRUE', and False for 'FALSE'.
        # Taken from: https://stackoverflow.com/a/56828137/748891

        # When unit testing, don't set the heroku config var, but do make
        #   the change to settings.
        if not self.sd.unit_testing:
            self.sd.write_output("  Setting DEBUG env var...")
            cmd = 'heroku config:set DEBUG=FALSE'
            output = self.sd.execute_subp_run(cmd)
            self.sd.write_output(output)
            self.sd.write_output("    Set DEBUG config variable to FALSE.")

        # Modify settings to use the DEBUG config variable.
        new_setting = "DEBUG = os.getenv('DEBUG') == 'TRUE'"
        msg_added = "    Added DEBUG setting for Heroku."
        msg_already_set = "    Found DEBUG setting for Heroku."
        self._add_heroku_setting(new_setting, msg_added, msg_already_set)


    def _configure_secret_key(self):
        """Use an env var to manage the secret key."""
        # Generate a new key.
        if self.sd.on_windows:
            # Non-alphanumeric keys have been problematic on Windows.
            new_secret_key = get_random_string(length=50,
                    allowed_chars='abcdefghijklmnopqrstuvwxyz0123456789')
        else:
            new_secret_key = get_random_secret_key()

        # Set the new key as an env var on Heroku.
        #   Skip when unit testing.
        if not self.sd.unit_testing:
            self.sd.write_output("  Setting new secret key for Heroku...")
            cmd = f"heroku config:set SECRET_KEY={new_secret_key}"
            output = self.sd.execute_subp_run(cmd)
            self.sd.write_output(output)
            self.sd.write_output("    Set SECRET_KEY config variable.")

        # Modify settings to use the env var's value as the secret key.
        new_setting = "SECRET_KEY = os.getenv('SECRET_KEY')"
        msg_added = "    Added SECRET_KEY setting for Heroku."
        msg_already_set = "    Found SECRET_KEY setting for Heroku."
        self._add_heroku_setting(new_setting, msg_added, msg_already_set)


    def _conclude_automate_all(self):
        """Finish automating the push to Heroku."""
        # Making this check here lets deploy() be cleaner.
        if not self.sd.automate_all:
            return

        self.sd.commit_changes()

        self.sd.write_output("  Pushing to heroku...")

        # Get the current branch name. Get the first line of status output,
        #   and keep everything after "On branch ".
        cmd = 'git status'
        git_status = self.sd.execute_subp_run(cmd)
        self.sd.write_output(git_status)
        status_str = git_status.stdout.decode()
        self.current_branch = status_str.split('\n')[0][10:]

        # Push current local branch to Heroku main branch.
        # This process usually takes a minute or two, which is longer than we
        #   want users to wait for console output. So rather than capturing
        #   output with subprocess.run(), we use Popen and stream while logging.
        # DEV: Note that the output of `git push heroku` goes to stderr, not stdout.
        self.sd.write_output(f"    Pushing branch {self.current_branch}...")
        if self.current_branch in ('main', 'master'):
            cmd = f"git push heroku {self.current_branch}"
        else:
            cmd = f"git push heroku {self.current_branch}:main"
        self.sd.execute_command(cmd)

        # Run initial set of migrations.
        self.sd.write_output("  Migrating deployed app...")
        if self.sd.nested_project:
            cmd = f"heroku run python {self.sd.project_name}/manage.py migrate"
        else:
            cmd = 'heroku run python manage.py migrate'
        output = self.sd.execute_subp_run(cmd)

        self.sd.write_output(output)

        # Open Heroku app, so it simply appears in user's browser.
        self.sd.write_output("  Opening deployed app in a new browser tab...")
        cmd = 'heroku open'
        output = self.sd.execute_subp_run(cmd)
        self.sd.write_output(output)


    def _summarize_deployment(self):
        """Manage all tasks related to generating and showing the friendly
        summary of the deployment.

        This does not take the place of the platform's official documentation.
          Instead, it gives the user a friendly entry into the platform's
          official documentation. It also gives them a brief summary of some
          followup steps they can take, for example making a second push, or
          changing the URL of the deployed app.
        """
        self._generate_summary()


    def _show_success_message(self):
        """After a successful run, show a message about what to do next."""

        # DEV:
        # - Say something about DEBUG setting.
        #   - Should also consider setting DEBUG = False in the Heroku-specific
        #     settings.
        # - Mention that this script should not need to be run again, unless
        #   creating a new deployment.
        #   - Describe ongoing approach of commit, push, migrate. Lots to consider
        #     when doing this on production app with users, make sure you learn.

        if self.sd.automate_all:
            # Show how to make future deployments.
            msg = dh_msgs.success_msg_automate_all(self.heroku_app_name,
                    self.current_branch)
        else:
            # Show steps to finish the deployment process.
            msg = dh_msgs.success_msg(self.sd.using_pipenv, self.heroku_app_name)

        self.sd.write_output(msg)


    # --- Utility methods ---

    def _check_current_heroku_settings(self, heroku_setting):
        """Check if a setting has already been defined in the heroku-specific
        settings section.
        """
        return any(heroku_setting in line for line in self.current_heroku_settings_lines)


    def _add_heroku_setting(self, new_setting, msg_added='',
            msg_already_set=''):
        """Add a new setting to the heroku-specific settings, if not already
        present.
        """
        already_set = self._check_current_heroku_settings(new_setting)
        if not already_set:
            with open(self.sd.settings_path, 'a') as f:
                self._prep_heroku_setting(f)
                f.write(f"\n    {new_setting}")
                self.sd.write_output(msg_added)
        else:
            self.sd.write_output(msg_already_set)


    def _prep_heroku_setting(self, f_settings):
        """Add a block for Heroku-specific settings, if it doesn't already
        exist.
        """
        if not self.found_heroku_settings:
            # DEV: Should check if `import os` already exists in settings file.
            f_settings.write("\nimport os")
            f_settings.write("\nif 'ON_HEROKU' in os.environ:")

            # Won't need to add these lines anymore.
            self.found_heroku_settings = True

    def _generate_summary(self):
        """Generate the friendly summary, which is html for now."""
        # Generate the summary file.
        path = self.sd.log_dir_path / 'deployment_summary.html'

        summary_str = "<h2>Understanding your deployment</h2>"
        path.write_text(summary_str)

        msg = f"\n  Generated friendly summary: {path}"
        self.sd.write_output(msg)


    # --- Methods called from simple_deploy.py ---

    def prep_automate_all(self):
        """Do intial work for automating entire process.

        Returns:
        - None if successful.
        """

        self.sd.write_output("  Running `heroku create`...")
        cmd = 'heroku create'
        output = self.sd.execute_subp_run(cmd)
        self.sd.write_output(output)


    def validate_platform(self):
        """Make sure the local environment and project supports deployment to
        Heroku.

        The returncode for a successful command is 0, so anything truthy means
          a command errored out.
        """
        # Make sure Heroku CLI is installed.
        cmd = 'heroku --version'
        output_obj = self.sd.execute_subp_run(cmd)
        if output_obj.returncode:
            raise CommandError(dh_msgs.cli_not_installed)
    


"""Handles all platform-agnostic aspects of the deployment process."""

# This is the command that's called to automate deployment.
# - It starts the process, and then dispatches to platform-specific helpers.
# - Each helper gets a reference to this command object.


import sys, os, re, subprocess, logging
from datetime import datetime
from pathlib import Path

from django.core.management.base import BaseCommand, CommandError
from django.conf import settings

from simple_deploy.management.commands.utils import deploy_messages as d_msgs
from simple_deploy.management.commands.utils import deploy_messages_heroku as dh_msgs
from simple_deploy.management.commands.utils import deploy_messages_platformsh as plsh_msgs
from simple_deploy.management.commands.utils import deploy_messages_flyio as flyio_msgs

from simple_deploy.management.commands.utils.deploy_heroku import HerokuDeployer
from simple_deploy.management.commands.utils.deploy_platformsh import PlatformshDeployer
from simple_deploy.management.commands.utils.deploy_flyio import FlyioDeployer


class Command(BaseCommand):
    """Perform the initial deployment of a simple project.
    Configure as much as possible automatically.
    """

    def add_arguments(self, parser):
        """Define CLI options."""

        # --- Platform-agnostic arguments ---

        parser.add_argument('--automate-all',
            help="Automate all aspects of deployment?",
            action='store_true')

        parser.add_argument('--platform', type=str,
            help="Which platform do you want to deploy to?",
            default='')

        # Allow users to skip logging.
        parser.add_argument('--no-logging',
            help="Do you want a record of simple_deploy's output?",
            action='store_true')

        # Allow users to use simple_deploy even with an unclean git status.
        parser.add_argument('--ignore-unclean-git',
            help="Run simple_deploy even with an unclean `git status` message.",
            action='store_true')

        # --- Platform.sh arguments ---

        # Allow users to set the deployed project name. This is the name that
        #   will be used by the platform, which may be different than the name
        #   used in the `startproject` command. See the Platform.sh script
        #   for use of this flag.
        parser.add_argument('--deployed-project-name', type=str,
            help="What name should the platform use for this project?\n(This is normally discovered automatically through inspection.)",
            default='')

        # Allow users to specify the region for a project when using --automate-all.
        parser.add_argument('--region', type=str,
            help="Which region do you want to deploy to?",
            default='us-3.platform.sh')

        # --- Developer arguments ---

        # If we're doing local unit testing, we need to avoid some network
        #   calls.
        parser.add_argument('--unit-testing',
            help="Used for local unit testing, to avoid network calls.",
            action='store_true')

        parser.add_argument('--integration-testing',
            help="Used for integration testing, to avoid confirmations.",
            action='store_true')


    def handle(self, *args, **options):
        """Parse options, and dispatch to platform-specific helpers."""
        self.stdout.write("Configuring project for deployment...")

        # Parse CLI options, and validate the set of arguments we've been given.
        self._parse_cli_options(options)
        self._validate_command()

        # Inspect system; we'll run some system commands differently on Windows.
        self._inspect_system()

        # Inspect project here. If there's anything we can't work with locally,
        #   we want to recognize that now and exit before making any changes
        #   to the project, and before making any remote calls.
        self._inspect_project()

        # Confirm --automate-all, if needed. Currently, this needs to happen before
        #   _validate_platform(), because fly_io takes action based on automate_all
        #   in _validate_platform().
        # Then build the platform-specifc deployer instance, and do platform-specific
        #   validation. 
        self._confirm_automate_all()
        self._validate_platform()

        # All validation has been completed. Make platform-agnostic modifications.
        # Start with logging.
        if self.log_output:
            self._start_logging()
            # Log the options used for this run.
            self.write_output(f"CLI args: {options}", write_to_console=False)

        # First action that could fail, but should happen after logging, is
        #   calling platform-specific prep_automate_all(). This usually creates
        #   an empty project on the target platform. This is one of the steps
        #   most likely to fail, so it should be called before other modifications.
        if self.automate_all:
            self.platform_deployer.prep_automate_all()

        self._add_simple_deploy_req()

        # During development, sometimes helpful to exit before calling deploy().
        # print('bye')
        # sys.exit()

        # All platform-agnostic work has been completed. Call platform-specific
        #   deploy() method.
        self.platform_deployer.deploy()


    # --- Internal methods; used only in this class ---

    def _parse_cli_options(self, options):
        """Parse cli options."""

        # Platform-agnostic arguments.
        self.automate_all = options['automate_all']
        self.platform = options['platform']
        # This is a True-to-disable option; turn it into a more intuitive flag?
        self.log_output = not(options['no_logging'])
        self.ignore_unclean_git = options['ignore_unclean_git']

        # Platform.sh arguments.
        self.deployed_project_name = options['deployed_project_name']
        self.region = options['region']

        # Developer arguments.
        self.unit_testing = options['unit_testing']
        self.integration_testing = options['integration_testing']


    def _validate_command(self):
        """Make sure the command has been called with a valid set of arguments.
        """
        if not self.platform:
            self.write_output(d_msgs.requires_platform_flag, write_to_console=False)
            raise CommandError(d_msgs.requires_platform_flag)


    def _validate_platform(self):
        """Find out which platform we're targeting, and instantiate the
        platform-specific deployer object. Also, call any necessary
        platform-specific validation and confirmation methods here.
        """
        if self.platform == 'heroku':
            self.write_output("  Targeting Heroku deployment...", skip_logging=True)
            self.platform_deployer = HerokuDeployer(self)
        elif self.platform == 'platform_sh':
            self.write_output("  Targeting platform.sh deployment...", skip_logging=True)
            self.platform_deployer = PlatformshDeployer(self)
            self.platform_deployer.confirm_preliminary()
        elif self.platform == 'fly_io':
            self.write_output("  Targeting Fly.io deployment...", skip_logging=True)
            self.platform_deployer = FlyioDeployer(self)
            self.platform_deployer.confirm_preliminary()
        else:
            error_msg = f"The platform {self.platform} is not currently supported."
            raise CommandError(error_msg)

        self.platform_deployer.validate_platform()


    def _confirm_automate_all(self):
        """If the --automate-all flag has been passed, confirm that the user
        really wants us to take these actions for them.
        """

        # Placing this test here makes handle() much cleaner.
        if not self.automate_all:
            return

        # Confirm the user knows exactly what will be automated; this
        #   message is specific to each platform.
        if self.platform == 'heroku':
            msg = dh_msgs.confirm_automate_all
        elif self.platform == 'platform_sh':
            msg = plsh_msgs.confirm_automate_all
        elif self.platform == 'fly_io':
            msg = flyio_msgs.confirm_automate_all

        self.write_output(msg, skip_logging=True)
        confirmed = self.get_confirmation(skip_logging=True)

        if confirmed:
            self.write_output("Automating all steps...", skip_logging=True)
        else:
            # Quit and have the user run the command again; don't assume not
            #   wanting to automate means they want to configure.
            self.write_output(d_msgs.cancel_automate_all, skip_logging=True)
            sys.exit()


    def _start_logging(self):
        """Set up for logging."""
        # Create a log directory if needed. Then create the log file, and 
        #   log the creation of the log directory if it happened.
        # In many libraries, one log file is created and then that file is
        #   appended to, and it's on the user to manage log sizes.
        # In this project, the user is expected to run simple_deploy
        #   once, or maybe a couple times if they make a mistake and it exits.
        # So, we should never have runaway log creation. It could be really
        #   helpful to see how many logs are created, and it's also simpler
        #   to review what happened if every log file represents a single run.
        # To create a new log file each time simple_deploy is run, we append
        #   a timestamp to the log filename.
        created_log_dir = self._create_log_dir()

        timestamp = datetime.now().strftime('%Y-%m-%d-%H%M%S')
        log_filename = f"simple_deploy_{timestamp}.log"
        verbose_log_path = self.log_dir_path / log_filename
        verbose_logger = logging.basicConfig(level=logging.INFO,
                filename=verbose_log_path,
                format='%(asctime)s %(levelname)s: %(message)s')

        self.write_output("Logging run of `manage.py simple_deploy`...",
                write_to_console=False)

        if created_log_dir:
            self.write_output(f"Created {self.log_dir_path}.")

        # Make sure we're ignoring sd logs.
        self._ignore_sd_logs()


    def _create_log_dir(self):
        """Create a directory to hold log files, if not already present.
        Returns True if created directory, False if directory was already
          present. Can't log from here, because log file has not been
          created yet.
        """
        self.log_dir_path = settings.BASE_DIR / Path('simple_deploy_logs')
        if not self.log_dir_path.exists():
            self.log_dir_path.mkdir()
            return True
        else:
            return False


    def _ignore_sd_logs(self):
        """Add log dir to .gitignore.
        Adds a .gitignore file if one is not found.
        """
        ignore_msg = "# Ignore logs from simple_deploy."
        ignore_msg += "\nsimple_deploy_logs/\n"

        # Assume .gitignore is in same directory as .git/ directory.
        # gitignore_path = Path(settings.BASE_DIR) / Path('.gitignore')
        gitignore_path = self.git_path / '.gitignore'
        if not gitignore_path.exists():
            # Make the .gitignore file, and add log directory.
            gitignore_path.write_text(ignore_msg)
            self.write_output("No .gitignore file found; created .gitignore.")
            self.write_output("Added simple_deploy_logs/ to .gitignore.")
        else:
            # Append log directory to .gitignore if it's not already there.
            # In r+ mode, a single read moves file pointer to end of file,
            #   setting up for appending.
            with open(gitignore_path, 'r+') as f:
                gitignore_contents = f.read()
                if 'simple_deploy_logs/' not in gitignore_contents:
                    f.write(f"\n{ignore_msg}")
                    self.write_output("Added simple_deploy_logs/ to .gitignore")


    def _strip_secret_key(self, line):
        """Strip secret key value from log file lines."""
        if 'SECRET_KEY:' in line:
            new_line = line.split('SECRET_KEY:')[0]
            new_line += 'SECRET_KEY: *value hidden*'
            return new_line
        else:
            return line


    def _inspect_system(self):
        """Find out if we're on Windows, so other methods can run system
        commands appropriately on each system. Note this is the user's system,
        not the host platform we're targeting.
        """
        if os.name == 'nt':
            self.on_windows = True
            self.use_shell = True
        else:
            self.on_windows = False
            self.use_shell = False


    def _inspect_project(self):
        """Find out everything we need to know about the project, before
        making any remote calls.

        - Determine project name.
        - Find .git/ directory.
        - Find out if this is a nested project or not.
        - Find significant paths: settings, project root, .git/ location.
        - Get the dependency management approach: requirements.txt, Pipenv, or
            Poetry.
        - Get the current requirements.          

        This method does the minimum introspection needed before making any
          remote calls. Anything that would cause us to exit before making the
          first remote call should be done here.
        """

        # Get project name. There are a number of ways to get the project
        #   name; for now we'll assume the root url config file has not
        #   been moved from the default location.
        # DEV: Use this code when we can require Python >=3.9.
        # self.project_name = settings.ROOT_URLCONF.removesuffix('.urls')
        self.project_name = settings.ROOT_URLCONF.replace('.urls', '')

        # Get project root, from settings.
        self.project_root = settings.BASE_DIR

        # Find .git location, and make sure there's a clean status.
        self._find_git_dir()
        self._check_git_status()

        self.settings_path = f"{self.project_root}/{self.project_name}/settings.py"

        self._get_dep_man_approach()
        self._get_current_requirements()


    def _find_git_dir(self):
        """Find .git location. Should be in BASE_DIR or BASE_DIR.parent.
        If it's in BASE_DIR.parent, this is a project with a nested
          directory structure.

        This method also sets self.nested_project. A nested project has the
          structure set up by:
           `django-admin startproject project_name`
        A non-nested project has manage.py at the root level, started by:
           `django-admin startproject .`
        This matters for knowing where manage.py is, and knowing where the
         .git dir is likely to be.
        """
        if Path(self.project_root / '.git').exists():
            self.git_path = Path(self.project_root)
            self.write_output(f"  Found .git dir at {self.git_path}.", skip_logging=True)
            self.nested_project = False
        elif (Path(self.project_root).parent / Path('.git')).exists():
            self.git_path = Path(self.project_root).parent
            self.write_output(f"  Found .git dir at {self.git_path}.", skip_logging=True)
            self.nested_project = True
        else:
            error_msg = "Could not find a .git/ directory."
            error_msg += f"\n  Looked in {self.project_root} and in {Path(self.project_root).parent}."
            raise CommandError(error_msg)


    def _check_git_status(self):
        """Make sure we're starting from a clean working state.
        We really want to encourage users to be able to easily undo
          configuration changes. This is especially true for automate-all.

        We do continue work if the only uncommitted change is adding 'simple_deploy'
          to INSTALLED_APPS.

        Returns:
            - None, if we can contine our work.
            - raises CommandError if we shouldn't continue.
        """

        if self.ignore_unclean_git:
            # People can override this check.
            return

        # If 'working tree clean' is in output of `git status`, we can definitely
        #   continue our work.
        cmd = "git status"
        output_obj = self.execute_subp_run(cmd)
        output_str = output_obj.stdout.decode()

        if "working tree clean" in output_str:
            return

        # `git status` indicated that uncommitted changes have been made.
        # Run `git diff`, and see if there are any changes beyond adding
        #   'simple_deploy' to INSTALLED_APPS.
        cmd = "git diff"
        output_obj = self.execute_subp_run(cmd)
        output_str = output_obj.stdout.decode()

        if not self._diff_output_clean(output_str):
            # There are uncommitted changes beyond just adding simple_deploy to
            #   INSTALLED_APPS, so we need to bail.
            error_msg = d_msgs.unclean_git_status
            if self.automate_all:
                error_msg += d_msgs.unclean_git_automate_all
            raise CommandError(error_msg)


    def _diff_output_clean(self, output_str):
        """Check output of `git diff`.
        If there are any lines that start with '- ', that indicates a deletion,
        and we'll bail.

        If there's more than one line starting with '+ ', that indicates more
        than one addition, and we'll bail.

        If there's only one addition, and it's 'simple_deploy' being added to
        INSTALLED_APPS, we can continue.

        Returns:
            - True if output of `git diff` is okay to continue.
            - False if output of `git diff` indicates we should bail.
        """
        num_deletions = output_str.count('\n- ')
        num_additions = output_str.count('\n+ ')

        if (num_deletions > 0) or (num_additions > 1):
            return False

        # There's only one addition. Check if it's anything other than
        #   simple_deploy being added to INSTALLED_APPS.
        # Note: This is not an overly specific test. We're not checking
        #   which file was changed, but there's no real reason someone would add
        #   'simple_deploy' by itself in another file.
        re_diff = r"""(\n\+{1}\s+[',"]simple_deploy[',"],)"""
        m = re.search(re_diff, output_str)
        if m:
            # The only addition was 'simple_deploy', we can move on.
            return True
        else:
            # There was a change, but it wasn't just 'simple_deploy'.
            # We should bail and have user look at their status.
            return False


    def _get_dep_man_approach(self):
        """Identify which dependency management approach the project uses.
        req_txt, poetry, or pipenv.
        """

        # In a simple project, I don't think we should find both Pipfile
        #   and requirements.txt. However, if there's both, prioritize Pipfile.
        self.using_req_txt = 'requirements.txt' in os.listdir(self.git_path)

        # DEV: If both req_txt and Pipfile are found, could just use req.txt.
        #      That's Heroku's prioritization, I believe. Address this if
        #      anyone has such a project, and ask why they have both initially.
        self.using_pipenv = 'Pipfile' in os.listdir(self.git_path)
        if self.using_pipenv:
            self.using_req_txt = False

        self.using_poetry = 'pyproject.toml' in os.listdir(self.git_path)
        if self.using_poetry:
            # Heroku does not recognize pyproject.toml, so we'll export to
            #   a requirements.txt file, and then work from that. This should
            #   not affect the user's local environment.
            cmd = 'poetry export -f requirements.txt --output requirements.txt --without-hashes'
            output = self.execute_subp_run(cmd)
            self.write_output(output, skip_logging=True)
            self.using_req_txt = True

        # Exit if we haven't found any requirements.
        if not any((self.using_req_txt, self.using_pipenv)):
            error_msg = f"Couldn't find any specified requirements in {self.git_path}."
            self.write_output(error_msg, write_to_console=False, skip_logging=True)
            raise CommandError(error_msg)


    def _get_current_requirements(self):
        """Get current project requirements, before adding any new ones.
        """
        if self.using_req_txt:
            # Build path to requirements.txt.
            self.req_txt_path = f"{self.git_path}/requirements.txt"

            # Get list of requirements, with versions.
            with open(f"{self.git_path}/requirements.txt") as f:
                requirements = f.readlines()
                self.requirements = [r.rstrip() for r in requirements]

        if self.using_pipenv:
            # Build path to Pipfile.
            self.pipfile_path = f"{self.git_path}/Pipfile"

            # Get list of requirements.
            self.requirements = self._get_pipfile_requirements()


    def _add_simple_deploy_req(self):
        """Add this project to requirements.txt."""
        # Since the simple_deploy app is in INCLUDED_APPS, it needs to be
        #   required. If it's not, Heroku will reject the push.
        # This step isn't needed for Pipenv users, because when they install
        #   django-simple-deploy it's automatically added to Pipfile.
        if self.using_req_txt:
            self.write_output("\n  Looking for django-simple-deploy in requirements.txt...")
            self.add_req_txt_pkg('django-simple-deploy')


    def _get_pipfile_requirements(self):
        """Get a list of requirements that are already in the Pipfile."""
        with open(self.pipfile_path) as f:
            lines = f.readlines()

        requirements = []
        in_packages = False
        for line in lines:
            # Ignore all lines until we've found the start of packages.
            #   Stop parsing when we hit dev-packages.
            if '[packages]' in line:
                in_packages = True
                continue
            elif '[dev-packages]' in line:
                # Ignore dev packages for now.
                break

            if in_packages:
                pkg_name = line.split('=')[0].rstrip()

                # Ignore blank lines.
                if pkg_name:
                    requirements.append(pkg_name)

        return requirements


    def _write_pipfile_pkg(self, package_name, version=""):
        """Write package to Pipfile."""

        with open(self.pipfile_path) as f:
            pipfile_text = f.read()

        if not version:
            version = '*'

        # Don't make ugly comments; make space to align comments.
        #   Align at column 30, so take away name length, and version spec space.
        tab_string = ' ' * (30 - len(package_name) - 5 - len(version))

        # Write package name right after [packages].
        #   For simple projects, this shouldn't cause any issues.
        #   If ordering matters, we can address that later.
        new_pkg_string = f'[packages]\n{package_name} = "{version}"{tab_string}# Added by simple_deploy command.'
        pipfile_text = pipfile_text.replace("[packages]", new_pkg_string)

        with open(self.pipfile_path, 'w') as f:
            f.write(pipfile_text)

        self.write_output(f"    Added {package_name} to Pipfile.")


    # --- Methods also used by platform-specific scripts ---

    def write_output(self, output_obj, log_level='INFO',
            write_to_console=True, skip_logging=False):
        """Write output to the appropriate places.
        Output may be a string, or an instance of subprocess.CompletedProcess.

        Need to skip logging before log file is configured.
        """

        # Extract the subprocess output as a string.
        if isinstance(output_obj, subprocess.CompletedProcess):
            # Assume output is either stdout or stderr.
            output_str = output_obj.stdout.decode()
            if not output_str:
                output_str = output_obj.stderr.decode()
        elif isinstance(output_obj, str):
            output_str = output_obj

        # Almost always write to console. Input from prompts is not streamed
        #   because user just typed it into the console.
        if write_to_console:
            self.stdout.write(output_str)

        # Log when appropriate. Log as a series of single lines, for better
        #   log file parsing. 
        if self.log_output and not skip_logging:
            for line in output_str.splitlines():
                # Strip secret key from any line that holds it.
                line = self._strip_secret_key(line)
                logging.info(line)


    def execute_subp_run_parts(self, cmd_parts):
        """This is similar to execute_subp_run(), but it receives a list of
        command parts rather than a string command. Having this separate method
        is cleaner than having nested if statements in execute_subp_run().

        Currently this is used to issue a git commit command, where running
          cmd.split() would split on the commit message.

        DEV: May want to make execute_subp_run() examine cmd that's received,
        and dispatch the work based on whether it receives a string or sequence.
        """
        if self.on_windows:
            cmd_string = ' '.join(cmd_parts)
            output = subprocess.run(cmd_string, shell=True, capture_output=True)
        else:
            output = subprocess.run(cmd_parts, capture_output=True)

        return output


    def execute_subp_run(self, cmd, check=False):
        """Execute subprocess.run() command.
        We're running commands differently on Windows, so this method
          takes a command and runs it appropriately on each system.

        The `check` parameter is included because some callers will need to 
          handle exceptions. For an example, see prep_automate_all() in
          deploy_platformsh.py. Most callers will only check whether returncode
          is nonzero, and not need to involve exception handling.

        Returns:
            - CompletedProcess instance
            - if check=True is passed, raises CalledProcessError. 
        """
        if self.on_windows:
            output = subprocess.run(cmd, shell=True, capture_output=True)
        else:
            cmd_parts = cmd.split()
            output = subprocess.run(cmd_parts, capture_output=True, check=check)

        return output


    def execute_command(self, cmd):
        """Execute command, and stream output while logging.
        This method is intended for commands that run long enough that we 
        can't use a simple subprocess.run(capture_output=True), which doesn't
        stream any output until the command is finished. That works for logging,
        but makes it seem as if the deployment is hanging. This is an issue
        especially on platforms like Azure that have some steps that take minutes
        to run.
        """

        # DEV: This only captures stderr right now.
        #   This is used for commands that run long enough that we don't
        #   want to use a simple subprocess.run(capture_output=True). Right
        #   now that's only the `git push heroku` call. That call writes to
        #   stderr; I'm not sure how to stream both stdout and stderr.
        #
        #     This will also be needed for long-running steps on other platforms,
        #   which may or may not write to stderr. Adding a parameter
        #   stdout=subprocess.PIPE and adding a separate identical loop over p.stdout
        #   misses stderr. Maybe combine the loops with zip()? SO posts on this
        #   topic date back to Python2/3 days.
        cmd_parts = cmd.split()
        with subprocess.Popen(cmd_parts, stderr=subprocess.PIPE,
            bufsize=1, universal_newlines=True, shell=self.use_shell) as p:
            for line in p.stderr:
                self.write_output(line)

        if p.returncode != 0:
            raise subprocess.CalledProcessError(p.returncode, p.args)


    def add_req_txt_pkg(self, package_name):
        """Add a package to requirements.txt, if not already present."""
        root_package_name = package_name.split('<')[0]

        # Note: This does not check for specific versions. It gives priority
        #   to any version already specified in requirements.
        pkg_present = any(root_package_name in r for r in self.requirements)

        if pkg_present:
            self.write_output(f"    Found {root_package_name} in requirements file.")
        else:
            with open(self.req_txt_path, 'a') as f:
                # Align comments, so we don't make req_txt file ugly.
                #   Version specs are in package_name in req_txt approach.
                tab_string = ' ' * (30 - len(package_name))
                f.write(f"\n{package_name}{tab_string}# Added by simple_deploy command.")

            self.write_output(f"    Added {package_name} to requirements.txt.")


    def add_pipenv_pkg(self, package_name, version=""):
        """Add a package to Pipfile, if not already present."""
        pkg_present = any(package_name in r for r in self.requirements)

        if pkg_present:
            self.write_output(f"    Found {package_name} in Pipfile.")
        else:
            self._write_pipfile_pkg(package_name, version)


    def get_confirmation(self, skip_logging=False):
        """Get confirmation for an action.
        This method assumes an appropriate message has already been displayed
          about what is to be done.
        This method shows a yes|no prompt, and returns True or False.
        """
        prompt = "\nAre you sure you want to do this? (yes|no) "
        confirmed = ''

        # If doing integration testing, always return True.
        if self.integration_testing:
            self.write_output(prompt, skip_logging=skip_logging)
            msg = "  Confirmed for integration testing..."
            self.write_output(msg, skip_logging=skip_logging)
            return True

        while True:
            self.write_output(prompt, skip_logging=skip_logging)
            confirmed = input()
            if confirmed.lower() in ('y', 'yes'):
                return True
            elif confirmed.lower() in ('n', 'no'):
                return False
            else:
                self.write_output("  Please answer yes or no.", skip_logging=skip_logging)


    def commit_changes(self):
        """Commit changes that have been made to the project.
        This should only be called when automate_all is being used.
        """
        if not self.automate_all:
            return

        self.write_output("  Committing changes...")
        cmd = 'git add .'
        output = self.execute_subp_run(cmd)
        self.write_output(output)
        # If we write this command as a string, the commit message will be split
        #   incorrectly.
        cmd_parts = ['git', 'commit', '-am', '"Configured project for deployment."']
        output = self.execute_subp_run_parts(cmd_parts)
        self.write_output(output)



"""A collection of messages used in deploy_heroku.py."""

# For conventions, see documentation in deploy_messages.py

from textwrap import dedent

from django.conf import settings


confirm_automate_all = """
The --automate-all flag means simple_deploy will:
- Run `heroku create` for you, to create a new Heroku project.
- Commit all changes to your project that are necessary for deployment.
  - These changes will be committed to the current branch, so you may want
    to make a new branch for this work.
- Push these changes to Heroku.
- Run the initial set of migrations to set up the remote database.
- Open your deployed project in a new browser tab.
"""

no_heroku_app_detected = """No Heroku app name has been detected.

- The simple_deploy command assumes you have already run 'heroku create'
  to start the deployment process.
- Please run 'heroku create', and then run
  'python manage.py simple_deploy' again.
- If you haven't already done so, you will need to install the Heroku CLI:
  https://devcenter.heroku.com/articles/heroku-cli
"""

cli_not_installed = """
In order to deploy to Heroku, you need to install the Heroku CLI.
  See here: https://devcenter.heroku.com/articles/heroku-cli
After installing the CLI, you can run simple_deploy again.
"""


# --- Dynamic strings ---
# These need to be generated in functions, to display information that's 
#   determined as the script runs.

def success_msg(using_pipenv, heroku_app_name):
    """Success message, when not using --automate-all flag."""

    # You can't use backslashes in f-strings, so this is the cleanest way I
    #   can add a pipenv line when needed.
    newline = '\n'

    msg = dedent(f"""

        --- Your project is now configured for deployment on Heroku. ---
        
        To deploy your project, you will need to:
        - Commit the changes made in the configuration process.
        - Push the changes to Heroku.
        - Migrate the database on Heroku.
        
        The following commands should finish your initial deployment:{newline + '        $ pipenv lock' if using_pipenv else ''}
        $ git add .
        $ git commit -am "Configured for Heroku deployment."
        $ git push heroku main
        $ heroku run python manage.py migrate
        
        After this, you can see your project by running 'heroku open'.
        Or, you can visit https://{heroku_app_name}.herokuapp.com.

    """)
    return msg


def success_msg_automate_all(heroku_app_name, current_branch):
    """Success message, when using --automate-all."""

    # Set correct command for pushing to heroku.
    if current_branch in ('main', 'master'):
        push_command = f"$ git push heroku {current_branch}"
    else:
        push_command = f"$ git push heroku {current_branch}:main"

    msg = dedent(f"""

        --- Your project should now be deployed on Heroku. ---

        It should have opened up in a new browser tab.
        - If you see the message "There's nothing here, yet"
          try waiting a moment and then refreshing your browser.
        - Sometimes when the process is automated there's a little lag
          before the project is fully deployed.
        - You can also visit your project at {heroku_app_name}.herokuapp.com.

        If you make further changes and want to push them to Heroku,
        commit your changes and then run the following command:
        {push_command}

        Also, if you haven't already done so you should review the
        documentation for Python deployments on Heroku at:
        - https://devcenter.heroku.com/categories/python-support
        - This documentation will help you understand how to maintain
          your deployment.

    """)
    return msg